/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * main.c
 * Copyright (C) 2014 WSID <jongsome@naver.com>
 * 
 * touchscreen-autodisabler is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * touchscreen-autodisabler is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

/* touchscreen-autodisabler
 *
 * This program disables and enables given device by watching proximity state
 * changes.
 * 
 * Wish: Man page
 *		 Needs code cleanup
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <signal.h>
#include <errno.h>

#include <stdbool.h>

#include "tad_util.h"
#include "tad_def.h"
#include "tad_app.h"
#include "tad_device.h"

/* **** Pre-initialization functions ******************************************/

void		tad_help					(void);
void		tad_version					(void);



/* **** Initialization / Finalization *****************************************/

void		tad_init					(void);
void		tad_interrupt				(void);



/* **** Callback functions ****************************************************/
void		tad_cb_toggle				(bool					enable,
						                 XIEventMask*			touchscreen);

void		tad_cb_device_print_machine	(TadEvent				event,
			                             void*					user_data);

void		tad_cb_device_print_human	(TadEvent				event,
			                             void*					user_data);

/* **** Static variables holds application state ******************************/


/*
 * main:
 * @argc:						commandline argument count
 * @argv: (array length=argc):  commandline arguments
 *
 * application entry point.
 * 
 * NOTE: this function may not return.
 * 
 * Returns: 0 on succes. error code, otherwises.
 */
int main (int argc, char *argv[])
{

	TadOption		opt_result;
	WatchCallback   watch_callback;

	tad_option_set(argc, argv, &opt_result);
	// if option says help or version, print it and exit.
	if (opt_result.help) {
		tad_option_help ();
		return TAD_EXIT_SUCCESS;
	}
	if (opt_result.version) {
		tad_version();
		return TAD_EXIT_SUCCESS;
	}

	// initialize base systems.
	

	
	if (opt_result.daemonize) tad_daemonize ();
	
	tad_init();
	
	tad_state_set((opt_result.clear_to_enable) ?
					TAD_TRANS_CLEAR_TO_ENABLE:
					TAD_TRANS_ZERO);
	
	tad_device_set_watch (opt_result.watch, opt_result.control);

	if (opt_result.print == TAD_PRINT_TYPE_HUMAN)
		watch_callback = (WatchCallback) tad_cb_device_print_human;
	else if (opt_result.print == TAD_PRINT_TYPE_MACHINE)
		watch_callback = (WatchCallback) tad_cb_device_print_machine;
	else 
		watch_callback = (WatchCallback) tad_transit;
	tad_device_watch (watch_callback, NULL);
	
	return TAD_EXIT_SUCCESS;
}




/* **** Function body - Pre-Initialization functions **************************/

/*
 * tad_version:
 * 
 * shows the version information.
 */
void
tad_version ( void )
{
	printf ("touchscreen_autodisabler - 0.9.5\t\tcopyright WSID\n");
}




/* **** Function body - Initialization / Finalization *************************/

/*
 * tad_init:
 * 
 * initializes application.
 * It opens display, and check for xinput, and get list of XDeviceInfo.
 * TODO: install signal handler to call tad_fini
 */
void
tad_init ( void )
{
	struct sigaction interrupt_sigact;

	
	DMSG("[INIT]Initializing\n");

	DMSG("[INIT]Installing SIGINT Signal Handler.\n");
	interrupt_sigact.sa_handler = (void(*)(int))tad_interrupt;
	sigemptyset (&interrupt_sigact.sa_mask);
	sigaddset   (&interrupt_sigact.sa_mask, SIGINT);
	interrupt_sigact.sa_flags = 0;
	
	if (sigaction (SIGINT, &interrupt_sigact, NULL) != 0) {
		perror ("touchscreen_autodisabler");
		ERR (TAD_EXIT_INIT_FAIL, "[INIT]Failed signal handler install\n");
	}
	
	DMSG ("[INIT]Locking.\n");
	tad_lock ();

	tad_device_init();

	tad_state_on_toggle ((TadToggleFunc)tad_cb_toggle, &mask_control);
}


/*
 * tad_interrupt:
 * 
 * finish and wrap up watching and controlling.
 */
void
tad_interrupt ( void )
{
	fputc ('\n', stderr);
	DMSG ("[SIGNAL]Handling SIGINT\n");
	exit (TAD_EXIT_SUCCESS);
}


/* **** Function body - callback functions ************************************/

/*
 * tad_cb_toggle:
 * @enable:			enable state.
 * @touchscreen:	event mask for touchscreen.
 * 
 * Callback function for tad_state_on_toggle().
 */
void
tad_cb_toggle (bool			enable,
			   XIEventMask*	touchscreen)
{
	tad_device_set_enabled (touchscreen, enable);
}

void
tad_cb_device_print_machine (TadEvent   event,
           					 void*		user_data)
{
	char	event_type;
	char	event_inout;
	char	state_c;
	char	touch_enable_c;
	
	tad_transit (event);
	
	event_type = ((event & 2) == 0) ? 'p' : 't';
	event_inout = ((event & 1) == 0) ? 'i' : 'o';

	state_c =			'0' + tad_state_get_state();
	touch_enable_c =	tad_state_get_touch_enabled () ? 't' : 'f';

	putchar (event_type);
	putchar (event_inout);
	putchar (state_c);
	putchar (touch_enable_c);
	putchar ('\n');
	fflush (stdout);
}

void
tad_cb_device_print_human   (TadEvent   event,
           					 void*		user_data)
{
	char*	event_type;
	char*	event_inout;
	
	tad_transit (event);
	
	event_type = ((event & 2) == 0) ? "pen  " : "touch";
	event_inout = ((event & 1) == 0) ? "in " : "out";

	printf ("%s %s: state: %d, touch enabled: %s\n",
	        event_type,
	        event_inout,
	        tad_state_get_state (),
	        BSTR (tad_state_get_touch_enabled ()) );
	fflush (stdout);
}