/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * tad_app.c
 * Copyright (C) 2014 WSID <jongsome@naver.com>
 * 
 * touchscreen-autodisabler is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * touchscreen-autodisabler is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <getopt.h>
#include <pwd.h>

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "tad_def.h"
#include "tad_util.h"
#include "tad_app.h"

/* macros *********************************************************************/
#define MEMBER(type, ptr, offset)   (*((type*)(((void*)(ptr)) + (offset))))

#define TAD_LOCK_FILE_USER_SUF  "/.touchscreen_autodisabler.pid"
#define TAD_HELP_OPT_SFORMAT "--%-18s -%c     %s\n"
#define TAD_HELP_OPT_LFORMAT "--%-18s        %s\n"


/* constants ******************************************************************/
const char* TadPrintTypeStr[] = {
	"no",
	"human",
	"machine"
};


const struct option long_opts[] = {
	{"help",			no_argument,		NULL,   OPTIONS_HELP},
	{"version",			no_argument,		NULL,   OPTIONS_VERSION},
	{"watch",			required_argument,	NULL,   OPTIONS_WATCH},
	{"control",			required_argument,	NULL,   OPTIONS_CONTROL},
	{"clear-to-enable", optional_argument,  NULL,   OPTIONS_CLEAR_TO_ENABLE},
	{"daemonize",		optional_argument,	NULL,   OPTIONS_DAEMONIZE},
	{"print",			required_argument,  NULL,   OPTIONS_PRINT},
	{NULL, 0, NULL, 0}
};

const char* opt_helps[] = {
	"Print this help message",
	"Print version information",
	"tablet device to watched.",
	"touchscreen device to be controlled.",
	"enable touchscreen after touch is clear (to prevent resting hand makes event).",
	"Run in background.",
	"prints events that happens"
};

const TadOptionItem opt_items[] = {
	{OPTION_VOID,   NULL,			offsetof(TadOption, help) },
	{OPTION_VOID,   NULL,			offsetof(TadOption, version) },
	{OPTION_STRING, NULL,			offsetof(TadOption, watch) },
	{OPTION_STRING, NULL,			offsetof(TadOption, control) },
	{OPTION_BOOL,   NULL,			offsetof(TadOption, clear_to_enable) },
	{OPTION_BOOL,   NULL,			offsetof(TadOption, daemonize) },
	{OPTION_ENUM,   TadPrintTypeStr,offsetof(TadOption, print) }
};


/*
 * trans_enable:
 * 
 * Touchscreen enabled state for each state of this application.
 * %true means touchscreen is enabled, and %false means touchscreen disabled.
 */
static const bool trans_enable[TAD_STATE_COUNT] = {
	true,   //TAD_STATE_ZERO
	false,  //TAD_STATE_PEN
	false,  //TAD_STATE_REST_PEN
	false,  //TAD_STATE_REST_STILL
	true	//TAD_STATE_TOUCHING
};

/*
 * trans_maps:
 * 
 * An array of transition maps (which is [#TadState][#TadEvent] -> #TadState)
 */
static const TadState
trans_maps[TAD_TRANS_COUNT][TAD_STATE_COUNT][TAD_EVENT_COUNT] = {
	{   //TAD_TRANS_ZERO
		{   //TAD_STATE_ZERO
			TAD_STATE_PEN,
			-1,
			TAD_STATE_ZERO,
			TAD_STATE_ZERO
		},
		{   //TAD_STATE_PEN
			-1,
			TAD_STATE_ZERO,
			TAD_STATE_PEN,
			TAD_STATE_PEN,
		}
	},
	{   //TAD_TRANS_CLEAR_TO_ENABLE
		{   //TAD_STATE_ZERO
			TAD_STATE_PEN,
			-1,
			TAD_STATE_TOUCHING, 
			-1,
		},
		{   //TAD_STATE_PEN
			-1,
			TAD_STATE_ZERO,
			TAD_STATE_REST_PEN,
			-1,
		},
		{   //TAD_STATE_REST_PEN
			-1,
			TAD_STATE_REST_STILL,
			-1,
			TAD_STATE_PEN,
		},
		{   //TAD_STATE_REST_STILL
			TAD_STATE_REST_PEN,
			-1,
			-1,
			TAD_STATE_ZERO,
		},
		{   //TAD_STATE_TOUCHING
			TAD_STATE_REST_PEN,
			-1,
			-1,
			TAD_STATE_ZERO
		}
	}
};


/* variables ******************************************************************/
static char*		lock_fname = NULL;
static int			lock_fd = -1;

/* static variables ***********************************************************/
static TadTransType		state_trans_type = TAD_TRANS_ZERO;
static TadState			state = TAD_STATE_ZERO;

static TadToggleFunc	toggle_func;
static void*			toggle_func_udata;

/* core application functions *************************************************/

/*
 * tad_lock_file_path:
 * 
 * retrives lock file path. Generally ~/.touchscreen_autodisabler.pid
 * 
 * Returns: lock file path for current user.
 */
char*
tad_lock_file_path (void)
{
	char*			result;

	char*			homedir;
	struct passwd*  pwd;
	
	homedir = getenv("HOME");
	if (homedir == NULL) {
		pwd = getpwuid(getuid());
		homedir = pwd->pw_dir;
	}
	
	result = malloc (sizeof (char) * strlen (homedir) +
	            	 sizeof (TAD_LOCK_FILE_USER_SUF));
	sprintf (result, "%s%s", homedir, TAD_LOCK_FILE_USER_SUF);

	return result;
}
	

/*
 * tad_lock:
 * 
 * create and lock a lock file as implementation of single instance system.
 * it uses fcntl() for locking. after locking, it will write pid.
 * if the file is locked, reads pid from it.
 */
void
tad_lock (void)
{
	struct flock	fl;

	char		buf[32];
	ssize_t			buf_len;
	int				pid_inst;

	lock_fname = tad_lock_file_path();
	DMSG ("[LOCK]Lock file path: %s\n", lock_fname);

	// Open up file.
	lock_fd = open (lock_fname, O_RDWR | O_CREAT, 0600);
	
	if (lock_fd < 0) {
		perror ("touchscreen_autodisabler");
		ERR (TAD_EXIT_CANNOT_LOCK, "[LOCK]Cannot open/create lock file.\n");
	}

	// prepares lock.

	fl.l_start = 0;
	fl.l_len = 0;
	fl.l_type = F_WRLCK;
	fl.l_whence = SEEK_SET;

	// On fail, reads pid from file.
	if (fcntl (lock_fd, F_SETLK, &fl) < 0) {
		buf_len = read (lock_fd, buf, 31);

		if (buf_len < 0)
			ERR (TAD_EXIT_CANNOT_LOCK, "[LOCK]Read error on lock file.\n");
		
		buf[buf_len] = '\0';

		if (! tad_str_is_integer (buf, &pid_inst)) 
			WRNG ("[LOCK]Lock file does not contains pid.\n");
			
		else
			printf ("another touchscreen_autodisabler instance is running.\n"
					"  pid: %d\n", pid_inst);
	
		close (lock_fd);
		exit (TAD_EXIT_SECOND_INST);
	}

	// On locking, writes pid of this instance.
	else {
		pid_inst = (int) getpid ();
		
		ftruncate (lock_fd, 0);
		buf_len = sprintf (buf, "%d", pid_inst);
		write (lock_fd, buf, buf_len);

		atexit (tad_unlock);
	}
}

/*
 * tad_unlock:
 * 
 * unlocks file lock and delete it.
 */
void
tad_unlock ( void )
{
	close (lock_fd);
	unlink (lock_fname);
	free (lock_fname);
}

/*
 * tad_daemonize:
 * 
 * daemonize running instance, and prints its pid.
 * 
 * NOTE: This will not return on parent side.
 */
void
tad_daemonize ( void )
{
	pid_t   pid;
	pid_t   sid;

	if (getppid() == 1) return;

	pid = fork ();

	if (pid < 0) {
		ERR (TAD_EXIT_DAEMON_FAIL, "[DAEMON]Child spawning failed.\n");
	}
	else if (pid > 0) {
		printf ("touchscreen_autodisabler instance spawned as daemon.\n"
		        "  pid: %d\n",
		        (int)pid);
		exit (TAD_EXIT_SUCCESS);
	}

	umask (0);
	sid = setsid();

	if (sid < 0) exit (TAD_EXIT_INIT_FAIL);
	if ((chdir("/")) < 0) exit (TAD_EXIT_INIT_FAIL);

	freopen ("/dev/null", "r", stdin);
	freopen ("/dev/null", "w", stdout);
	freopen ("/dev/null", "w", stderr);
}


/* command line option processing *********************************************/
/*
 * tad_option:
 * @argc:			argc from main()
 * @argv:			argv from main()
 * @result: (out):  option parse result for this application.
 *
 * parses commandline options with getopt_long().
 *
 * Returns: %TRUE if succesfull. %FALSE on otherwise.
 */
void
tad_option_set (int					argc,
            	char**				argv,
            	TadOption*			result)
{
	char					opt_short[64];
	int						opt_res;
	int						opt_index;

	const struct option*	opt;
	char*					str_p;

	bool					bool_value;
	int						enum_value;

	// Initial values.
	result->watch = NULL;
	result->control = NULL;
	
	result->help = false;
	result->version = false;
	result->daemonize = false;
	
	result->clear_to_enable = false;
	result->print = 0;
	
	// Generate short option string.
	for (str_p = opt_short, opt = long_opts;  opt->name != NULL;  opt++) {
		if (isalpha (opt->val)) {
			*(str_p++) = (char)opt->val;
			if (opt->has_arg != no_argument) {
				*(str_p++) = ':';
				if (opt->has_arg == optional_argument)
					*(str_p++) = ':';
			}
		}
	}
	*str_p = '\0';
	DMSG("[ARG]Generated short option is: %s\n", opt_short);

	opterr = 0;
	
	// Parse options.
	do {
		opt_index = -1;
		opt_res = getopt_long (argc, argv, opt_short, long_opts, &opt_index);
		DMSG("[ARG]opt_res: %c(%d), optopt: %c, optarg: %s\n",
		     (char)opt_res, opt_res, (char)optopt, optarg);

		// Determine opt_res.
		if (opt_res == '?') opt_res = optopt;
		else if (opt_res == -1) break;

		// Gets opt_index if option was given in short form.
		if (opt_index == -1)
			for (opt_index = 0; long_opts[opt_index].name != NULL; opt_index++)
				if (long_opts[opt_index].val == opt_res) break;
		
		if (long_opts[opt_index].name == NULL)
			ERR (TAD_EXIT_ARG_WRONG, "[ARG] Wrong commandline argument.\n");

		// Process option by its type.
		switch (opt_items[opt_index].type) {
			case OPTION_VOID:
				MEMBER(bool, result, opt_items[opt_index].offset) = true;
				break;

				
			case OPTION_BOOL:
				if (optarg == NULL)
					bool_value = true;
				
				else if (! tad_str_is_bool (optarg, &bool_value))
					ERR (TAD_EXIT_ARG_WRONG_PARAM,
					     "[ARG] %s: yes or no expected: %s\n",
					     long_opts[opt_index].name,
					     optarg);
				
				MEMBER(bool, result, opt_items[opt_index].offset) = bool_value;
				break;

				
			case OPTION_ENUM:
				if (optarg == NULL) {
					ERR (TAD_EXIT_ARG_WRONG_PARAM,
					     "[ARG] %s: needs parameter for this option.\n",
					     long_opts[opt_index].name);
				}
				else {
					enum_value = tad_strv_get_index ((const char**)opt_items[opt_index].type_data,
					                        optarg);

					if (enum_value == -1)
						ERR (TAD_EXIT_ARG_WRONG_PARAM,
						     "[ARG] %s: cannot be parameter: %s",
						     long_opts[opt_index].name,
						     optarg);
					MEMBER(int, result, opt_items[opt_index].offset) = enum_value;
				}
				break;

				
			case OPTION_STRING:
				if (optarg == NULL) {
					ERR (TAD_EXIT_ARG_WRONG_PARAM,
					     "[ARG] %s: needs parameter for this option.\n",
					     long_opts[opt_index].name);
				}
				else
					MEMBER(char*, result, opt_items[opt_index].offset) = optarg;
				break;
		}
	}
	while (opt_res != -1);

	if (result->help || result->version)
		return;
	
	else if (result->watch == NULL) {
		ERR (TAD_EXIT_ARG_MISSING,
		     "[ARG]Watched device(-w, --watch) is required.\n");
	}
	else if (result->control == NULL) {
		ERR (TAD_EXIT_ARG_MISSING,
		     "[ARG]Controlled device(-c, --control) is required.\n");
	}
}

/*
 * tad_option_to_string:
 * @option:	option to convert string.
 * 
 * converts option into string equivalent.
 * 
 * Returns: string to create @option back.
 */
char*
tad_option_to_string (TadOption*	option)
{
	char*	result;
	char*	result_p;
	size_t	len;
	
	len = strlen (option->watch) + 3;
	len += strlen (option->control) + 3 + 1;
	
	if (option->clear_to_enable) len += 2 + 1;
	if (option->daemonize) len += 2 + 1;
	
	result = (char*) malloc (len * sizeof(char));
	
	result_p = result + sprintf (result,
								 "-w %s -c %s",
								 option->watch,
								 option->control);
								 
	if (option->clear_to_enable)
		result_p += sprintf (result, " -e");
	
	if (option->daemonize)
		result_p += sprintf (result, " -d");
	
	return result;
}

/*
 * tad_option_help:
 * 
 * shows help message of this application.
 */
void
tad_option_help ( void )
{
	int i;
	int j;
	
	printf ("touchscreen_autodisabler:\n"
	        "  automatic disable and enables as tablet actives,\n"
	        "  so that hand can rest on touchscreen.\n"
	        "\n"
	        "  touchscreen_autodisabler -w \"device1\" -c \"device2\"\n"
	        "  device1: tablet device to be watched.\n"
	        "  device2: touchscreen device to be controlled.\n"
	        "\n");

	for (i = 0; long_opts [i].name != NULL; i++) {
		if (isalpha((char)(long_opts [i].val)))
			printf (TAD_HELP_OPT_SFORMAT,
			        long_opts[i].name,
			        (char)(long_opts[i].val),
			        opt_helps[i]);
		else
			printf (TAD_HELP_OPT_LFORMAT,
			        long_opts[i].name,
			        opt_helps[i]);

		if (opt_items[i].type == OPTION_ENUM) {
			printf ("%30spossible values:", "");
			for (j = 0; ((char**)opt_items[i].type_data)[j] != NULL; j++) {
				if (j != 0) fputs (", ", stdout);
				fputs (((char**)opt_items[i].type_data)[j], stdout);
			}
			fputc('\n', stdout);
		}
	}
	printf ("\n"
	        "If you want put names with multi words, quote the name so that\n"
	        "commandline send the name as whole.\n");
}


/* functions body - state management */

/*
 * tad_state_set:
 * @trans_type: the type that state would transit.
 * 
 * Sets transit type before using tad_state_* function groups.
 */
void
tad_state_set (TadTransType   trans_type)
{
	state_trans_type = trans_type;
}

/*
 * tad_state_get_state:
 * 
 * gets current state.
 * 
 * Returns: current state.
 */
TadState		
tad_state_get_state (void)
{
	return state;
}

/*
 * tad_state_get_touch_enabled:
 * 
 * gets whether touchscreen supposed to be enabled or not.
 * 
 * Returns: %TRUE if touchscreen supposed to be enabled.
 */
bool
tad_state_get_touch_enabled (void)
{
	return trans_enable [state];
}

/*
 * tad_state_on_toggle:
 * @func:		state toggle callback.
 * @user_data:  user data for @func
 * 
 * register @func as state toggle function, which is invoked when touchscreen
 * enabled state changes.
 * 
 * Returns: previous callback.
 */
TadToggleFunc   
tad_state_on_toggle (TadToggleFunc  func,
				     void*			user_data)
{
	TadToggleFunc prev_func;

	prev_func = toggle_func;

	toggle_func = func;
	toggle_func_udata = user_data;
	return prev_func;
}

/*
 * tad_transit:
 * @stimulus:   stimulus to be given.
 * 
 * change state according to @stimulus.
 * Also touchscreen state also changed by its new state.
 */
void
tad_transit (TadEvent	stimulus)
{
	bool	prev_enable;
	bool	new_enable;

	prev_enable = trans_enable[state];
	state = trans_maps [state_trans_type][state][stimulus];
	new_enable = trans_enable[state];

	if (prev_enable != new_enable) toggle_func (new_enable, toggle_func_udata);
}