/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * tad_device.c
 * Copyright (C) 2014 WSID <jongsome@naver.com>
 * 
 * touchscreen-autodisabler is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * touchscreen-autodisabler is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <fcntl.h>
#include <sys/time.h>

#include <stdbool.h>
#include <stdarg.h>
#include <string.h>
#include <inttypes.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/extensions/XInput.h>
#include <X11/extensions/XInput2.h>

#include "tad_def.h"
#include "tad_util.h"
#include "tad_device.h"

int				xi_opcode = 0;
int				xi_event_code = 0;
int				xi_error_code = 0;

XIEventMask		masks[2];
// macro	mask_watched: masks[0]
// macro	mask_control: masks[1]


/* display:
 * opened display that currently running on.
 */
Display*		display = NULL;

/*
 * window:
 * window to grab device.
 */
Window			window = NULL;

/*
 * devinfo_list:
 * list of device information.
 */
XDeviceInfo*	devinfo_list = NULL;

/*
 * devinfo_list_count:
 * number of devices.
 */
int				devinfo_list_count = 0;




/* **** Device functions ******************************************************/

/*
 * tad_device_init:
 * 
 * initialize Xlib usage and other things.
 */
void
tad_device_init	(void) {
	display = XOpenDisplay(getenv ("DISPLAY"));
	window = DefaultRootWindow (display);
	
	if (!XQueryExtension(display, "XInputExtension",&xi_opcode, &xi_event_code, &xi_error_code))
		ERR (TAD_EXIT_INIT_FAIL, "[INIT]XInput Extension is not availiable!\n");

	devinfo_list = XListInputDevices(display, &devinfo_list_count);

	atexit (tad_device_fini);
}

/*
 * tad_device_fini:
 * 
 * finalize Xlib usage and other things.
 */
void
tad_device_fini (void) {
	XFreeDeviceList (devinfo_list);
	XCloseDisplay (display);
}

/*
 * tad_device_set_watch:
 * @watched: a name of watched tablet device.
 * @control: a name of controlled touchscreen device.
 * 
 * Set device to be watched and to be controlled.
 * 
 * Once this is set, they can be watched by tad_device_watch().
 */
void
tad_device_set_watch (const char*	watched,
				      const char*	control)
{
	XID				watched_id;
	XID				control_id;

	watched_id = tad_device_id_from_args (watched);
	control_id = tad_device_id_from_args (control);
	
	tad_mask_set (&mask_watched, watched_id, XI_PropertyEvent, 0);
	tad_mask_set (&mask_control, control_id, XI_RawTouchBegin, XI_RawTouchEnd, 0);
}


/*
 * tad_device_watch:
 * @callback:						a callback to be invoked for each events.
 * @user_data: (closure callback):  a supplement data.
 * 
 * Watches the devices that was setted from tad_device_set_watch() and invoke
 * callback function for each event.
 * 
 * Note: this will block forever.
 */
void
tad_device_watch (WatchCallback callback,
				  void*			user_data)
{
	XEvent					event;
	XGenericEventCookie*	ev_cookie;
	XIDeviceEvent*			ev_dev;

	int						conn_fd;
	fd_set					conn_fd_set;

	TadEvent				stimulus;

	ev_cookie = &event.xcookie;
	conn_fd = ConnectionNumber (display);
	
	DMSG ("[WATCH]Registering event type to listen.\n");
	XISelectEvents (display, window, masks, 2);
	
	DMSG ("[WATCH]Watching events.\n");

	while (true) {
		FD_ZERO (&conn_fd_set);
		FD_SET  (conn_fd, &conn_fd_set);

		if (!XPending (display))
			if (!select (conn_fd + 1, &conn_fd_set, NULL, NULL, NULL))
				continue;

		while (XPending(display)) {
			XNextEvent(display, &event);

			if ((XGetEventData (display, ev_cookie)) &&
				(ev_cookie->type == GenericEvent) &&
				(ev_cookie->extension == xi_opcode))
			{
				ev_dev = (XIDeviceEvent*)ev_cookie->data;
				DMSG ("[WATCH%d]Event type %d\n",
					  (int)ev_dev->deviceid,
					  (int)ev_dev->evtype);
				stimulus = tad_device_event_transform (ev_dev);
				if (stimulus != TAD_EVENT_INVAILD)
					callback (stimulus, user_data);
			}
		}
	}
}


/*
 * tad_mask_set:
 * @mask:				mask to set.
 * @device_id:			device id.
 * @first_eventtype:	first event type.
 * @...:				additional event type. terminated with 0
 * 
 * sets #XIEventMask structure in one line.
 * just for convenience.
 */
void
tad_mask_set	(XIEventMask*   mask,
				 XID			device_id,
				 int			first_eventtype, ...)
{
	va_list event_list;
	
	int		event;

	va_start (event_list, first_eventtype);
	
	mask->deviceid = device_id;
	mask->mask_len = XIMaskLen ( XI_LASTEVENT );
	mask->mask = (unsigned char*) calloc (mask->mask_len, sizeof (unsigned char));

	if (first_eventtype != 0) {
		XISetMask (mask->mask, first_eventtype);

		for (event = va_arg (event_list, int);
		     event != 0;
		     event = va_arg (event_list, int))

			XISetMask (mask->mask, event);
	}
	va_end (event_list);
}

/*
 * tad_device_id_from_args:
 * @name:   device name or xid in string.
 * 
 * gets xinput device id from user arguments.
 * 
 * Returns: XID of device.
 */
XID
tad_device_id_from_args (const char*	name)
{
	const XDeviceInfo*	device_info;
	int					device_id_int;
	XID					device_id;
	
	const XDeviceInfo* devinfo_pos;
	const XDeviceInfo* devinfo_end;
	
	device_info = NULL;

	devinfo_end = devinfo_list + devinfo_list_count;

	if (! tad_str_is_integer (name, (int *)&device_id_int)) {
		DMSG("[DEV]Looking up %s\n", name);
		for (devinfo_pos = devinfo_list;
			 devinfo_pos < devinfo_end;
			 devinfo_pos ++) {
			if (strcmp(name, devinfo_pos->name) == 0) {
				device_info = devinfo_pos;
				break;
			}
		}
		if (device_info == NULL)
			ERR (TAD_EXIT_WRONG_DEV, "[DEV]Device not found: %s\n", name);
		
		device_id = device_info -> id;
	}
	else {
		DMSG("[DEV]%s is number - treat as Device ID (%d)\n", name, device_id_int);
		device_id = (XID)device_id_int;
	}

	return device_id;
}

/*
 * tad_device_set_enabled:
 * @device_id:  XID of device to disable.
 * @enabled:	enabled status to set.
 * 
 * Enables or disables device.
 */
void
tad_device_set_enabled (XIEventMask*	mask,
		                bool			enabled)
{
	if (!enabled) {
		DMSG ("[DEV%d]Grabbing events.", (int)mask->deviceid);

		XIGrabDevice (display,
			          mask->deviceid,
			          window,
			          CurrentTime,
			          None,
			          GrabModeAsync,
			          GrabModeAsync,
			          true,
			          mask);
	}
	else {
		DMSG ("[DEV%d]Ungrabbing touching event.", (int)mask->deviceid);
		XIUngrabDevice (display,
		                mask->deviceid,
		                CurrentTime);
	}
}

/*
 * tad_device_event_transform:
 * @event: event to be transformed.
 * 
 * Transforms #XIDeviceEvent structure into #TadEvent.
 * 
 * Returns: a corresponding #TadEvent
 */
TadEvent
tad_device_event_transform (XIDeviceEvent*  event)
{
	static Atom		atom_wac_ser_id = 0;
	static int		touch_count = 0;
	
	bool			proximity;		// proximity side
	
	Atom			type_return;	// for wacom side.
	int				format_return;  // ''

	unsigned long	num_item;		// ''
	unsigned long   bytes_after;	// ''
	unsigned char*  data;			// ''

	uint32_t* data_prox;			// ''

	XIPropertyEvent* prop_event;	// ''
	

	if (atom_wac_ser_id == 0)
		atom_wac_ser_id = XInternAtom (display, "Wacom Serial IDs", false); 

	proximity = false;

	
	// tablet (watched)
	
	if (event->deviceid == mask_watched.deviceid) {
		if (event->evtype == XI_PropertyEvent) {
			prop_event = (XIPropertyEvent*)event;

			if (prop_event->property == atom_wac_ser_id) {
				DMSG ("[DEV%d]Processing property event.\n", event->deviceid);
				// Gets serial ids.
				XIGetProperty (display,
							   event->deviceid,
							   atom_wac_ser_id,
							   0,
							   256,
							   false,
							   XA_INTEGER,
							   &type_return,
							   &format_return,
							   &num_item,
							   &bytes_after,
							   &data);

				data_prox = ((uint32_t*)data) + 3;
				proximity = (*data_prox != 0);
				DMSG ("[DEV%d] data prox: %" PRIu32 "\n", (int)event->deviceid, *data_prox);
				DMSG ("[DEV%d] proximity: %s\n",
					  (int)event->deviceid,
					  proximity ? "In" : "Out");
				XFree (data);
			}
			else {
				DMSG ("[DEV%d]Uninterested property change discarded.: %d\n",
					  (int)prop_event->deviceid,
					  (int)prop_event->property);
				return TAD_EVENT_INVAILD;
			}
		}
		else {
			DMSG ("[DEV%d]Uninterested event discarded.: %d\n",
				  (int)event->deviceid,
				  (int)event->evtype);
			return TAD_EVENT_INVAILD;
		}
		DMSG ("[DEV]Proximity %s\n", proximity ? "In" : "Out");
		return proximity ? TAD_EVENT_PEN_IN : TAD_EVENT_PEN_OUT;
	}
	

	// touchscreen (control)
	
	else if (event->deviceid == mask_control.deviceid) {
		// Touch begin
		if (event->evtype == XI_RawTouchBegin) {
			DMSG ("[DEV]Touch count increased.\n");
			touch_count++;
			if (touch_count == 1) {
				DMSG ("[DEV]Touch begin!\n");
				return TAD_EVENT_TOUCH_IN;
			}
			return TAD_EVENT_INVAILD;
		}
		// Touch end
		else if (event->evtype == XI_RawTouchEnd){
			DMSG ("[DEV]Touch count decreased.\n");
			touch_count--;
			if (touch_count == 0) {
				DMSG ("[DEV]Touch end!\n");
				return TAD_EVENT_TOUCH_OUT;
			}
			else if (touch_count < 0) {
				WRNG ("[DEV] Touch count less 0.\n");
				touch_count = 0;
			}
			return TAD_EVENT_INVAILD;
		}
		
		else {
			DMSG ("[DEV%d]Uninterested event discarded.\n", (int)event->deviceid);
			return TAD_EVENT_INVAILD;
		}
	}


	else {
		DMSG ("[DEV%d]Event from uninterested device discarded.\n",
		      (int)event->deviceid);
		return TAD_EVENT_INVAILD;
	}

	WRNG ("[DEV]This statement should not reached!.\n");
	return TAD_EVENT_INVAILD;
}
